from ._client import Client, Connection
from ._types import (
    Message,
    MessageId,
    MessageInfo,
    NewsItem,
    NewsItemId,
    NewsItemInfo,
    Person,
    Pupil,
    PupilId,
    ReplyMessage,
)

__all__ = [
    'Client',
    'Connection',
    'Message',
    'MessageId',
    'MessageInfo',
    'NewsItem',
    'NewsItemId',
    'NewsItemInfo',
    'Person',
    'Pupil',
    'PupilId',
    'ReplyMessage',
]



import textwrap
from dataclasses import dataclass
from datetime import datetime
from typing import Iterable, List, NamedTuple, NewType, Optional

from bs4 import BeautifulSoup

PupilId = NewType('PupilId', str)
MessageId = NewType('MessageId', int)
NewsItemId = NewType('NewsItemId', int)


class Pupil(NamedTuple):
    id: PupilId
    name: str


@dataclass
class Person:
    name: str
    id: Optional[int] = None
    type: Optional[str] = None


@dataclass
class MessageInfo:
    id: MessageId
    origin: str  # URL of the system this message is from
    pupil_id: PupilId
    subject: str
    last_timestamp: datetime
    folder: str
    sender: Person
    reply_count: int
    is_unread: bool

    def __str__(self) -> str:
        return (
            f'Message {self.id:6}: '
            f'{self.last_timestamp:%Y-%m-%d %H:%M} '
            f'"{self.subject}"'
            f'{f" (+{self.reply_count})" if self.reply_count else ""}')


class _MessageWithBody:
    timestamp: datetime
    sender: Person
    body: str

    def __str__(self) -> str:
        return self.to_text()

    def to_text(self, width: int = 70) -> str:
        return (
            f'{self.get_header_lines()}'
            f'\n'
            f'{self.get_cleaned_body_text(width)}')

    def get_header_lines(self) -> str:
        return (
            f'Date: {self.timestamp}\n'
            f'From: {self.sender.name}\n'
        )

    def get_cleaned_body_text(self, width: int = 70) -> str:
        body_text = BeautifulSoup(self.body, features='lxml').get_text()
        return '\n\n'.join(
            '\n'.join(textwrap.wrap(
                line.replace('\xa0', ' ').rstrip(),
                width=width))
            for line in body_text.splitlines())


@dataclass
class ReplyMessage(_MessageWithBody):
    timestamp: datetime
    sender: Person
    body: str


@dataclass
class Message(_MessageWithBody, MessageInfo):
    timestamp: datetime
    recipients: List[Person]
    body: str
    replies: List[ReplyMessage]

    @classmethod
    def from_info_and_attrs(
            cls,
            info: MessageInfo,
            timestamp: datetime,
            recipients: Iterable[Person],
            body: str,
            replies: Iterable[ReplyMessage] = (),
    ) -> 'Message':
        return cls(
            id=info.id,
            origin=info.origin,
            pupil_id=info.pupil_id,
            subject=info.subject,
            timestamp=timestamp,
            last_timestamp=info.last_timestamp,
            folder=info.folder,
            sender=info.sender,
            reply_count=info.reply_count,
            is_unread=info.is_unread,
            recipients=list(recipients),
            body=body,
            replies=list(replies),
        )

    def get_header_lines(self) -> str:
        return f'Subject: {self.subject}\n' + super().get_header_lines()


@dataclass
class NewsItemInfo:
    id: NewsItemId
    origin: str  # URL of the system this message is from
    pupil_id: PupilId
    subject: str
    timestamp: Optional[datetime]
    is_unread: bool

    def __str__(self) -> str:
        date = self.timestamp.date() if self.timestamp else "????-??-??"
        return f'News {self.id}: {date} "{self.subject}"'


@dataclass
class NewsItem(_MessageWithBody, NewsItemInfo):
    timestamp: datetime
    sender: Person
    body: str

    @classmethod
    def from_info_and_attrs(
            cls,
            info: NewsItemInfo,
            *,
            timestamp: datetime,
            sender: Person,
            body: str,
    ) -> 'NewsItem':
        return cls(
            id=info.id,
            origin=info.origin,
            pupil_id=info.pupil_id,
            subject=info.subject,
            is_unread=info.is_unread,
            timestamp=timestamp,
            sender=sender,
            body=body,
        )

    def get_header_lines(self) -> str:
        return f'Subject: {self.subject}\n' + super().get_header_lines()



from typing import Any, Optional, Protocol, Sequence, TypeVar, Union

from .element import Tag

_Str_T = TypeVar('_Str_T', str, bytes, covariant=True)


class _Readable(Protocol[_Str_T]):
    def read(self, n: int = ...) -> _Str_T: ...


_Str = Union[str, bytes]


class BeautifulSoup(Tag):
    def __init__(
            self,
            markup: Union[_Str, _Readable[Any]] = ...,
            features: Optional[Union[str, Sequence[str]]] = ...,
            #
            # The following parameters are unused in this code base
            # and therefore left undefined here:
            #
            # builder: Optional[...] = ...,
            # parse_only: Optional[...] = ...,
            # from_encoding: Optional[...] = ...,
            # exclude_encodings: Optional[...] = ...,
            # element_classes: Optional[...] = ...,
            # **kwargs: object,
    ) -> None: ...



from typing import (
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    TypeVar,
    Union,
    overload,
)

_T = TypeVar('_T')
_MatchAgainst = Union[str, bool]  # incomplete (e.g. Callable, Iterable, ...)


class SoupStrainer:
    ...


class ResultSet(List[_T]):
    def __init__(
            self,
            source: SoupStrainer,
            result: Iterable[_T] = ...,
    ) -> None: ...


class PageElement:
    ...


class Tag(PageElement):
    name: str

    next_element: Optional[Tag]
    next_sibling: Optional[Tag]
    parent: Optional[Tag]
    previous_element: Optional[Tag]
    previous_sibling: Optional[Tag]

    @overload
    def get(self, key: str) -> Optional[str]: ...

    @overload
    def get(self, key: str, default: _T) -> Union[str, _T]: ...

    def __setitem__(self, key: str, value: str) -> None: ...

    def get_text(
            self,
            separator: str = ...,
            strip: bool = ...,
            types: Optional[Sequence[type]] = ...
    ) -> str: ...

    @property
    def text(self) -> str: ...

    def find(
            self,
            name: Optional[str] = ...,
            attrs: Mapping[str, _MatchAgainst] = ...,
            recursive: bool = ...,
            text: Optional[str] = ...,
            **kwargs: _MatchAgainst,
    ) -> Optional['Tag']: ...

    def find_all(
            self,
            name: str = ...,
            attrs: Mapping[str, _MatchAgainst] = ...,
            recursive: bool = ...,
            text: Optional[str] = ...,
            limit: Optional[int] = ...,
            **kwargs: _MatchAgainst,
    ) -> ResultSet['Tag']: ...

    def find_next(
            self,
            name: Optional[str] = ...,
            attrs: Mapping[str, _MatchAgainst] = ...,
            text: Optional[str] = ...,
            limit: Optional[int] = ...,
            **kwargs: _MatchAgainst,
    ) -> Optional['Tag']: ...

    def find_previous(
            self,
            name: Optional[str] = ...,
            attrs: Mapping[str, _MatchAgainst] = ...,
            text: Optional[str] = ...,
            limit: Optional[int] = ...,
            **kwargs: _MatchAgainst,
    ) -> Optional['Tag']: ...

    def select(
            self,
            selector: str,
            # namespaces=None, limit=None, **kwargs,
    ) -> List['Tag']: ...

    def select_one(
            self,
            selector: str,
            # namespaces=None, **kwargs,
    ) -> Optional['Tag']: ...

    def replace_with(self, replace_with: Union[str, 'Tag']) -> 'Tag': ...

    def __iter__(self) -> Iterator['Tag']: ...



from typing import Any, Mapping, Optional

import bs4
import requests


class _Response(requests.Response):
    @property
    def soup(self) -> Optional[bs4.BeautifulSoup]: ...


class Browser:
    def __init__(
            self,
            # TODO: Replace Any with Session
            session: Optional[Any] = ...,
            soup_config: Optional[Mapping[str, Any]] = ...,
            # TODO: Replace Any with BaseAdapter below
            requests_adapters: Optional[Mapping[str, Any]] = ...,
            raise_on_404: bool = ...,
            user_agent: Optional[str] = ...,
    ) -> None: ...

    def request(
            self,
            method: str,
            url: str,
            headers: Mapping[str, str] = ...,
            # *args, **kwargs,
    ) -> _Response: ...

    def get(
            self,
            url: str,
            headers: Mapping[str, str] = ...,
            # *args, **kwargs,
    ) -> _Response: ...

    def post(
            self,
            url: str,
            headers: Mapping[str, str] = ...,
            # *args, **kwargs,
    ) -> _Response: ...



import binascii

from bs4.element import Tag

URL_PART = '/cdn-cgi/l/email-protection#'


def unmangle_emails(element: Tag) -> None:
    _replace_email_links(element)
    _replace_email_texts(element)


def _replace_email_links(element: Tag) -> None:
    for a_elem in element.find_all('a'):
        href = a_elem.get('href', '')
        (_before, _sep, after_url_part) = href.partition(URL_PART)
        if after_url_part:
            a_elem['href'] = 'mailto:' + _unmangle(after_url_part)


def _replace_email_texts(element: Tag) -> None:
    for elem in element.find_all(attrs={'class': '__cf_email__'}):
        hex_string = elem.get('data-cfemail')
        if hex_string:
            email = _unmangle(hex_string)
            elem.replace_with(email)


def _unmangle(hex_string: str) -> str:
    values = binascii.unhexlify(hex_string)
    first_val = values[0]
    return ''.join(chr(x ^ first_val) for x in values[1:])



from typing import Optional

import bs4

from .browser import Browser, _Response
from .form import Form


class StatefulBrowser(Browser):
    def absolute_url(self, url: str) -> str: ...

    def open(
            self,
            url: str,
            # *args: ...,
            # **kwargs: ...,
    ) -> _Response: ...

    def open_relative(
            self,
            url: str,
            #*args, **kwargs,
    ) -> _Response: ...

    def get_url(self) -> Optional[str]: ...

    def select_form(
            self,
            selector: str = ...,
            nr: int = ...,
    ) -> Form: ...

    def __setitem__(self, name: str, value: str) -> None: ...

    def submit_selected(
            self,
            # btnName=None,
            # update_state=True,
            # *args,
            # **kwargs,
    ) -> _Response: ...

    def get_current_page(self) -> Optional[bs4.BeautifulSoup]: ...
